/*
 * $RCSfile: route_xlate.c,v $
 *
 */
#include "xsi_main.h"
#include <sys/socket.h>
#include <net/route.h>
#include "ip.h"


static void
IPhost_2_sockaddr(IPhost * iph, struct sockaddr * sa)
{
  sa->sa_family = AF_INET;
  sa->sa_data[2] = iph->a;
  sa->sa_data[3] = iph->b;
  sa->sa_data[4] = iph->c;
  sa->sa_data[5] = iph->d;
}

static void
sockaddr_2_IPhost(IPhost * iph, struct sockaddr * sa)
{
  iph->a = sa->sa_data[2];
  iph->b = sa->sa_data[3];
  iph->c = sa->sa_data[4];
  iph->d = sa->sa_data[5];
}

/*
 * translate from BSD route objects (struct rtentry)
 * to xkernel route info (IProuteInfo)
 */
void
route_BSD_2_Xk(IProuteInfo * dest, struct rtentry * src)
{
  IPhost mask = { 0xFF, 0, 0, 0 };

  sockaddr_2_IPhost(&dest->net, &src->rt_dst);
  sockaddr_2_IPhost(&dest->gw, &src->rt_gateway);
  if (dest->net.a & 0x80) {	/* class B or more */
    mask.b = 0xFF;
    if (dest->net.a & 0x40) {	/* class C or more */
      mask.c = 0xFF;
      if (dest->net.a & 0x20) {	/* class D or more */
        mask.d = 0xFF;		/* a guess */
        if (dest->net.a & 0x10) {
		/* class E address */
        }
      }
    }
  }
  dest->mask = mask; 
  dest->metric = 1;
  dest->ttl = 0;
}

/*
 * translate from xkernel route info (IProuteInfo) to
 * BSD route objects (struct rtentry)
 */
void
route_Xk_2_BSD(struct rtentry * dest, IProuteInfo * src)
{
  int isahost = (src->mask.d != 0);

  IPhost_2_sockaddr(&src->net, &dest->rt_dst);
  IPhost_2_sockaddr(&src->gw, &dest->rt_gateway);
  dest->rt_flags = RTF_UP | (isahost ? RTF_HOST : 0);
  dest->rt_hash = 0;
  dest->rt_refcnt = 0;
  dest->rt_use = 0;
  dest->rt_ifp = 0;
}
