/*
 * $RCSfile: host_list.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: host_list.c,v $
 * Revision 1.1  1995/10/17 22:24:02  ho
 * Initial revision
 *
 * Revision 1.2  1995/04/28  00:21:08  dorgival
 * Minor syntax changes to make the IRIX compiler happy
 *
 * Revision 1.1  1995/04/24  18:48:15  dorgival
 * Initial revision
 *
 * Revision 1.1  1995/04/05  17:52:37  dorgival
 * Initial revision
 *
 *
 */

#include "xkernel.h"
#include "km.h"
#include "grpmem.h"
#include "host_list.h"

extern int tracegrpmemp;

void host_list_dump ( host_list* hl );

void
host_list_dump ( host_list* hl )
{
  int i;

  fprintf(stderr,"%d hosts:", HOST_LIST_SIZE(hl));
  for ( i=0;i<HOST_LIST_SIZE(hl);++i) {
    fprintf(stderr," \"%s\"",ipHostStr(HOST_LIST_HOST(hl,i)));
  }
  fprintf(stderr,"\n");
}

void
host_list_init( host_list* hl )
{
  HOST_LIST_SIZE(hl)  = 0;
  HOST_LIST_HOSTS(hl) = (LOCALHOSTTYPE*) xMalloc( GRPMEM_MAX_PARTS * sizeof(LOCALHOSTTYPE));
  bzero( (char*)HOST_LIST_HOSTS(hl), GRPMEM_MAX_PARTS * sizeof(LOCALHOSTTYPE) );
}

void
host_list_destroy( host_list* hl )
{
  HOST_LIST_SIZE(hl)  = 0;
  xFree((char*)HOST_LIST_HOSTS(hl));
  HOST_LIST_HOSTS(hl) = NULL;
}

void
host_list_add( hl, h )
  host_list*     hl;
  LOCALHOSTTYPE* h;
{
  int i;

  xTrace1(grpmemp, TR_NEVER, "host_list_add \"%s\"", ipHostStr(h));
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
  for (i=0;i<HOST_LIST_SIZE(hl);++i)
    if ( same_host( h, HOST_LIST_HOST(hl,i) ) )
      break;
  if ( i == GRPMEM_MAX_PARTS )
    xTrace0(grpmemp, TR_ERRORS, "GRPMEM: host_list_add exceeded group size");
  else if ( i == HOST_LIST_SIZE(hl) ) {
    xTrace0(grpmemp, TR_NEVER, "adding host");
    HOST_LIST_ADD(hl,i,h);
    HOST_LIST_SIZE(hl)++;
  }
  else
    xTrace0(grpmemp, TR_NEVER, "host already in list");
  xTrace1(grpmemp, TR_NEVER, "host_list_add \"%s\" - returning", ipHostStr(h));
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
}

void
host_list_del( hl, h )
  host_list*     hl;
  LOCALHOSTTYPE* h;
{
  int i;

  xTrace1(grpmemp, TR_NEVER, "host_list_del \"%s\"", ipHostStr(h));
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
  for (i=0;i<HOST_LIST_SIZE(hl);++i)
    if ( same_host( h, HOST_LIST_HOST(hl,i) ) )
      break;
  if ( i<HOST_LIST_SIZE(hl) ) {
    xTrace1(grpmemp,TR_NEVER, "deleting host from position %d", i );
    for ( --(HOST_LIST_SIZE(hl)); i<HOST_LIST_SIZE(hl); ++i )
      bcopy((char*)HOST_LIST_HOST(hl,i+1),(char*)HOST_LIST_HOST(hl,i),sizeof(LOCALHOSTTYPE));
  }
  else {
    xTrace0(grpmemp ,TR_SOFT_ERRORS, "GRPMEM: host_list_del->host not in list");
  }
  bzero((char*)HOST_LIST_HOST(hl,HOST_LIST_SIZE(hl)),sizeof(LOCALHOSTTYPE));
  xTrace1(grpmemp, TR_NEVER, "host_list_del \"%s\" - returning", ipHostStr(h));
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
}

LOCALHOSTTYPE*
host_list_find( hl, h )
  host_list*     hl;
  LOCALHOSTTYPE* h;
{
  int i;

  xTrace1(grpmemp, TR_NEVER, "host_list_find \"%s\"", ipHostStr(h));
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
  for (i=0;i<HOST_LIST_SIZE(hl);++i)
    if ( same_host( h, HOST_LIST_HOST(hl,i) ) )
      break;
  if (i<HOST_LIST_SIZE(hl)) {
    xTrace2(grpmemp, TR_NEVER,"found \"%s\" in position %d",
		     ipHostStr(HOST_LIST_HOST(hl,i)), i );
    return HOST_LIST_HOST(hl,i);
  }
  else {
    xTrace0(grpmemp, TR_NEVER,"host not found");
    return NULL;
  }
}

void
host_list_foreach( hl, func, arg1, arg2 )
  host_list* hl;
  foreach_func_t* func;
  void*      arg1;
  void*      arg2;
{
  int i;

  xTrace0(grpmemp, TR_NEVER, "host_list_foreach");
  xIfTrace(grpmemp, TR_NEVER)
    host_list_dump( hl );
  for ( i=0; i< HOST_LIST_SIZE(hl); ++i ){
    if ( ((int)func( HOST_LIST_HOST(hl,i), arg1, arg2 )) < 0 )
      return ;
  }
  xTrace0(grpmemp, TR_NEVER, "host_list_foreach - returning");
  return;
}

void
host_list_clear( hl )
  host_list* hl;
{
  HOST_LIST_SIZE(hl) = 0;
}
