/*     
 * $RCSfile: ipsec_simplespi.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994,1993,1991,1990  Arizona Board of Regents
 *
 *
 *
 */


#include "xkernel.h"
#include "ip.h"
#include "km.h"
#include "ipsecspi.h"
#include "ipsec.h"
#include "saidmgr.h"
#include "ipsec_simplespi_i.h"

int	traceipsecspip;


#ifdef __STDC__
static void		ipsecspi_getproc_sessn(XObj);
static void		ipsecspi_getproc_protl(XObj);
static xkern_return_t	ipsecspiCloseSessn( XObj );
static int		ipsecspiControlProtl( XObj, int, char *, int ) ;
static xkern_return_t	ipsecspiDemux( XObj, XObj, Msg * );
static XObj		ipsecspiOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecspiOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecspiOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	ipsecspiOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	ipsecspiPop( XObj, XObj, Msg *, void * );
static xmsg_handle_t	ipsecspiPush( XObj, Msg * );
static long 		ipsecspiGetSaid(void *, char *, long, void *);
static void 		ipsecspiPutSaid(void *, char *, long, void *);


#else
static void		ipsecspi_getproc_sessn();
static void		ipsecspi_getproc_protl();
static xkern_return_t	ipsecspiCloseSessn();
static int		ipsecspiControlProtl() ;
static xkern_return_t	ipsecspiDemux();
static XObj		ipsecspiOpen();
static xkern_return_t	ipsecspiOpenDisable();
static xkern_return_t	ipsecspiOpenEnable();
static xkern_return_t	ipsecspiOpenDone();
static xkern_return_t	ipsecspiPop();
static xmsg_handle_t	ipsecspiPush();
static long 		ipsecspiGetSaid();
static void 		ipsecspiPutSaid();

#endif /* __STDC__ */

LOCALHOSTTYPE Any_Host = {-1, -1, -1, -1};

/*
 * ipsecspi_init
 */
void
ipsecspi_init(self)
    XObj self;
{
    PSTATE	*pstate;
    XObj llp;
    int i;
    
    xTrace0(ipsecspip, TR_GROSS_EVENTS, "IPSEC init");
    xAssert(xIsProtocol(self));
    
    ipsecspi_getproc_protl(self);
    pstate = X_NEW(PSTATE);
    bzero((char *)pstate, sizeof(PSTATE));
    self->state = (VOID *)pstate;

    /* 
      active map is used for open/demux.
      lookup based on <lsaid,hls>.

      */
    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));

    pstate->passivemap = mapCreate(PASSIVE_MAP_SIZE, sizeof(PassiveId));
 

    for(i=0; i < self->numdown; i++)
    {
	llp = xGetDown(self,i);
	if (strcmp(llp->name,"km") == 0)
		pstate->km = llp;
	else
		pstate->transport = llp;
    }
    xTrace0(ipsecspip, TR_GROSS_EVENTS, "IPSEC init done");
}

/*
 * ipsecspiOpen
 */
static XObj
ipsecspiOpen( self, hlpRcv, hlpType, p )
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    ActiveId	key;
    XObj s;
    PSTATE	*pstate = (PSTATE *)(self->state);
    xTrace0(ipsecspip, TR_MAJOR_EVENTS, "IPSEC open");

    key.lsaid = *((XSaid *)partPop(p[2]));
    key.lls = xOpen(self, self, pstate->transport, p);
    key.prot = relProtNum(hlpType, self);
    if (key.lls != ERR_XOBJ) {

      xTrace2(ipsecspip, TR_DETAILED, "ipsecspi open said %u lls %x ",
	      key.lsaid, key.lls);

      if (mapResolve(pstate->activemap, &key, &s) == XK_SUCCESS)
	{
	  xTrace0(ipsecspip, TR_MAJOR_EVENTS, "IPSEC open: reusing session");

	  return s;
	}
      else {
    	SSTATE	*sstate;
    	Part	kmp[1];
    	KMA_MAPPING kmma;	
    	XObj	kms;
	IPhost  myhost;

	xTrace2(ipsecspip, TR_MAJOR_EVENTS, "IPSEC open: new session lsaid %u lls %x",
		 key.lsaid,key.lls);
    	s = xCreateSessn(ipsecspi_getproc_sessn,
		     hlpRcv, hlpType, self, 1, &key.lls);
    	sstate = (SSTATE *)xMalloc(sizeof(SSTATE));
    	bzero((char *)sstate, sizeof(SSTATE));
    	s->state = (char *)sstate;
        s->binding = mapBind(pstate->activemap, &key, s);

    	sstate->lsaid = key.lsaid;

	if ((xControl(key.lls, GETMYHOST, (char*) &myhost, LOCALHOSTLEN))==XK_FAILURE)
	{	
		xTrace0(ipsecspip, TR_MAJOR_EVENTS, "IPSEC open: can't retrieve my host address");
		return ERR_XOBJ;
	}
     	partInit(kmp,1);
     	partPush(kmp[0], (char *)&key.lsaid,sizeof(XSaid));
     	partPush(kmp[0], (char *)&myhost,sizeof(IPhost));	
     	kms = xOpen(self, self, pstate->km , kmp);
        if (kms == ERR_XOBJ) return ERR_XOBJ;
     	xControl(kms, KM_RESOLVE, (char *)&kmma, sizeof(kmma));
    	xClose(kms); 

	sstate->rsaid = kmma.remote_said;
	sstate->ihdr.said = kmma.remote_said;
	sstate->ihdr.prot = key.prot;
	xTrace1(ipsecspip, TR_MAJOR_EVENTS, "IPSEC open: new session rsaid %u",
		 sstate->rsaid);
	return s;
        }
    }
    return ERR_XOBJ;
}

static xkern_return_t
ipsecspiOpenDone( self, lls, llp, hlpType)
     XObj self, lls, llp, hlpType;
{
  /* PSTATE *ps; */

  xTrace0(ipsecspip, TR_EVENTS, "ipsecspi: opendone called, doing nothing");
  return XK_SUCCESS;
}

static xkern_return_t
ipsecspiOpenEnable(s, h, htype, p)
     XObj s, h, htype;
     Part *p;
{
  PassiveId key;
  Enable *en;
  LOCALHOSTTYPE * rh;

  xTrace0(ipsecspip, TR_EVENTS, "ipsecspi: openenable called");
  
  key.algo = *((XAlgorithm *)partPop(p[2]));
  key.prot = relProtNum(htype, s);
  xTrace2(ipsecspip, TR_EVENTS, "ipsecspi: openenable algorithm %u for protocol %u",
	key.algo, key.prot);
  rh = (LOCALHOSTTYPE *)partPop(p[REMOTE_PART]);
  if (rh)
  {
        if (rh == ANY_HOST)
        {
		key.h = Any_Host;
  		partPush(p[REMOTE_PART],ANY_HOST, 0);
        }
        else
        {
		key.h = *rh;
  		partPush(p[REMOTE_PART],(char*) &key.h, LOCALHOSTLEN);
        }
  }
  else
  {
	return XK_FAILURE;
  }

  if ((mapResolve(((PSTATE *)(s->state))->passivemap, &key, &en)==XK_FAILURE)
      ||
      (en->hlpRcv != h))
    {
      en = X_NEW(Enable);
      bzero((char *)en, sizeof(Enable));
      en->hlpRcv = h;
      en->hlpType = htype;
      en->binding = mapBind(((PSTATE *)(s->state))->passivemap, &key, en);
      xTrace2(ipsecspip, TR_EVENTS, "ipsecspi: openenable added to map %u %u",
	key.algo, key.prot);
      xTrace4(ipsecspip, TR_EVENTS, "ipsecspi address is %u %u %u %u",
	key.h.a, key.h.b, key.h.c, key.h.d);
    }
  en->rcnt++;
  return(xOpenEnable(s, s, ((PSTATE *)(s->state))->transport,p));
}


/*
 * ipsecspiPush
 */
static xmsg_handle_t
ipsecspiPush(s, msg)
     XObj s;
     Msg *msg;
{
    SSTATE	*sstate;
    
    xAssert(xIsSession(s));
    sstate = (SSTATE *) s->state;
    
    xTrace2(ipsecspip, TR_EVENTS, "ipsecspi push lsaid %u rsaid %u", sstate->lsaid, sstate->rsaid);

    msgPush(msg, ipsecspiPutSaid, &sstate->ihdr, IPSEC_HDR_NETSIZE, 0);
    return xPush(xGetDown(s,0), msg);
  }

static xkern_return_t
ipsecspiPop( s, lls, m, hdr)
  Msg *m;
  XObj s, lls;
  VOID *hdr;
{
  return xDemux(s, m);
}

static long ipsecspiGetSaid(dst, src, len, arg)
     void *dst;
     char *src;
     long len;
     void *arg;
{
  ipsecHdr *dest = (ipsecHdr *)dst;
  short ntord;
  xAssert(len>=IPSEC_HDR_NETSIZE);
  bcopy(src, (char*) &dest->said, SAID_NETSIZE);
  bcopy(src+SAID_NETSIZE, (char *)&ntord, PROT_NETSIZE);
  dest->prot = ntohs(ntord);
  return IPSEC_HDR_NETSIZE;
}

static void ipsecspiPutSaid(src, dst, len, arg)
     void *src;
     char *dst;
     long len;
     void *arg;
{
  ipsecHdr *sorc = (ipsecHdr *)src;
  short htord;
  xAssert(len==IPSEC_HDR_NETSIZE);
  bcopy((char *) &sorc->said, dst, SAID_NETSIZE);
  htord = htons(sorc->prot);
  bcopy((char *)&htord, dst+SAID_NETSIZE, PROT_NETSIZE);
}

/*
 * ipsecspiDemux
 */
static xkern_return_t
ipsecspiDemux(self, transport_s, dg)
    XObj self;
    XObj transport_s;
    Msg *dg;
{
    XObj        s;
    ipsecHdr	ihdr;
    ActiveId	key;
    PassiveId	pkey;
    PSTATE	*pstate = (PSTATE *) self->state;

    xTrace2(ipsecspip, TR_EVENTS,
	    "IPSECSPI demux called with datagram of len %d, lls %x", msgLen(dg),
	    transport_s);

    msgPop(dg, ipsecspiGetSaid, &ihdr,  IPSEC_HDR_NETSIZE, 0);
    key.lls = transport_s;
    key.lsaid = ihdr.said;
    key.prot = ihdr.prot;
    if ( mapResolve(pstate->activemap, &key, &s) == XK_FAILURE )
    {
      Enable *e;
      Part	kmp[1];
      KMA_MAPPING kmma;	
      XObj	kms;
      SSTATE	*sstate;
      IPhost	myhost;

      if ((xControl(key.lls, GETMYHOST, (char*) &myhost, LOCALHOSTLEN))==XK_FAILURE)
      {	
	xTrace0(ipsecspip, TR_MAJOR_EVENTS, "IPSEC demux: can't retrieve my host address");
	return XK_FAILURE;
      }
      partInit(kmp,1);
      partPush(kmp[0],(char *) &key.lsaid,sizeof(XSaid));	
      partPush(kmp[0],(char *) &myhost,sizeof(IPhost));	
      kms = xOpen(self, self, pstate->km , kmp);
      if (kms == ERR_XOBJ) return XK_FAILURE;
      xControl(kms, KM_RESOLVE, (char *) &kmma, sizeof(kmma));
      xClose(kms);

      pkey.algo = kmma.local_algorithm;
      pkey.prot = key.prot;
      xControl(transport_s, GETPEERHOST, (char*) &pkey.h, LOCALHOSTLEN);
      
      if ( mapResolve(pstate->passivemap, &pkey, &e ) == XK_FAILURE )
	{
	    	xTrace2(ipsecspip, TR_EVENTS, "ipsecspi no passive entry for  %u %u ",
		        pkey.algo, pkey.prot);
	    	xTrace4(ipsecspip, TR_EVENTS, "ipsecspi address is %u %u %u %u",
			pkey.h.a, pkey.h.b, pkey.h.c, pkey.h.d);
		pkey.h = Any_Host;
      	    	if ( mapResolve(pstate->passivemap, &pkey, &e ) == XK_FAILURE )
	    	{
	    		xTrace2(ipsecspip, TR_EVENTS, "ipsecspi no passive session found for %u %u",
		        pkey.algo, pkey.prot);
	    		xTrace4(ipsecspip, TR_EVENTS, "ipsecspi address is %u %u %u %u",
				pkey.h.a, pkey.h.b, pkey.h.c, pkey.h.d);
	   	 	return XK_FAILURE;
		}
	}
     	s = xCreateSessn(ipsecspi_getproc_sessn,
		     e->hlpRcv, e->hlpType, self, 1, &key.lls);
    	sstate = (SSTATE *)xMalloc(sizeof(SSTATE));
    	bzero((char *)sstate, sizeof(SSTATE));
    	s->state = (char *)sstate;
        s->binding = mapBind(pstate->activemap, &key, s);
    	sstate->lsaid = key.lsaid;

	sstate->rsaid = kmma.remote_said;
	sstate->ihdr.said = kmma.remote_said;
	sstate->ihdr.prot = key.prot;
        xDuplicate(transport_s);
        xOpenDone(e->hlpRcv, s, self);
      }
     else
      xTrace2(ipsecspip, TR_DETAILED, "ipsecspi demux using existing session for %x %x",
	      key.lsaid, key.lls);

     if (((SSTATE *)(s->state))->flags & IP_PS_HACK_FLAG
	|| ((PSTATE *)(xMyProtl(s)->state))->flags & IP_PS_HACK_FLAG)
     {  
      	IPpseudoHdr *ipphdr = msgGetAttr( dg, 0);
      	if (ipphdr)
	{  
	  long len;

	  len = msgLen(dg);
	  ipphdr->len = htons(ntohs(ipphdr->len) - IPSEC_HDR_NETSIZE);
	  xTrace4( ipsecspip, TR_FULL_TRACE,
		   "IPSECSPI Pop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len %d  dlen %d",
		   ipphdr, ntohs(ipphdr->len), len, len-IPSEC_HDR_NETSIZE); 
	  ipphdr->prot = ((SSTATE *)(s->state))->ihdr.prot; 
	 }
     }
     return xPop(s, transport_s, dg, (void *)0);
}


/*
 * ipsecspiControlProtl
 */
static int
ipsecspiControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
    xAssert(xIsProtocol(self)||xIsSession(self));
    
    switch (opcode) {
      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;

	    if (len < sizeof(max)) return 0; 
	    if (xControl(xGetDown(self, 0), opcode, (char *) &max, sizeof(max)) <= 0)
	      {
		return 0; 
	      }
	    max -= IPSEC_HDR_NETSIZE;
	    *(int *)buf = max;
	    return(sizeof(max));
	}
      case SAIDMGR_GETLSAID:
        { 
		if (xIsSession(self))
		{	
			return ((SSTATE *)self->state)->lsaid;
		}
        }
      case GETPARTICIPANTS:
        { 
		if (xIsSession(self))
		{	
    			Part   p[3];
			Part   p2[2];
    			char * ebuf;
    			int    elen = len;
			int    x;
			ebuf = xMalloc(elen);
			if (xControl(xGetDown(self,0),GETPARTICIPANTS, ebuf, elen) <= 0)
				return -1; 
			partInit(p,3);
			partInternalize(p2,ebuf);
			for(x=0; x < partLen(p2); x++)
			{
				elen = partStackTopByteLen(p2[x]);
				partPush(p[x],partPop(p2[x]),elen);
			} 
    			partPush(p[2],&(((SSTATE *)self->state)->lsaid),sizeof(XSaid));	
			partExternalize(p, buf, &len);
			xFree(ebuf);
			return len;
		}
        }

      case GETPEERHOST :
      case IPSEC_GETPEERHOST :
	{
	  int val;
	  checkLen(len, LOCALHOSTLEN);
	  val = xControl(xGetDown(self,0), GETPEERHOST, buf, len);
	  return val;
	}

      case GETMYHOST :
      case IPSEC_GETMYHOST:
	{
	  int val;
	  checkLen(len, LOCALHOSTLEN);
	  val = xControl(xGetDown(self,0), GETMYHOST, buf, len);
	  return val;
	}
	
      case IP_GETPSEUDOHDR:
	{
	  int val;
	  SSTATE *ss = (SSTATE *)self->state;
	  
	  xAssert(xIsSession(self));
	  val = xControl(xGetDown(self,0), opcode, buf, len);
	  ((IPpseudoHdr *)(buf))->prot = ss->ihdr.prot; 
	  return val;
      }
	break;

      case IP_PSEUDOHDR:
	/* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PSTATE *)(self->state))->flags |= 1;
	    xTrace0(ipsecspip, TR_EVENTS,
	      "IPSECSPI: ControlProtl: Turning on IPpseudoheader length fixup.");  }
/* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too.  **** */
/* **** Code for this control-op should be immediately above default: **** */      
      default:
        {
    		if (buf && len) bzero(buf, len);   /* prevent info going down */
    		return xControl(((PSTATE *)(self->myprotl->state))->transport, opcode, buf, len);
        }
     }
}

static xkern_return_t
ipsecspiCloseSessn(s)
     XObj s;
{
  if (s->rcnt == 0)
  {
  	xTrace0(ipsecspip, TR_FULL_TRACE, "ipsecspiCloseSessn called");
  	mapRemoveBinding(((PSTATE *)(s->myprotl->state))->activemap, s->binding);
  	xClose(xGetDown(s,0));
  	xFree((char *)(s->state));
  }
  return XK_SUCCESS;
}

static xkern_return_t
  ipsecspiOpenDisable( s, h1, h2, p)
  XObj s, h1, h2;
  Part *p;
{
  PassiveId key;
  Enable *en;
  LOCALHOSTTYPE *rh;

  xTrace0(ipsecspip, TR_FULL_TRACE, "ipsecspiOpenDisable called");
  /* only supports dissables with exactly the same participants as enables */
  
  key.algo = *((XAlgorithm *)partPop(p[2]));
  key.prot = relProtNum(h2, s);
  rh = (LOCALHOSTTYPE *)partPop(p[REMOTE_PART]);
  if (rh)
  {
        if (rh == ANY_HOST)
        {
		key.h = Any_Host;
  		partPush(p[REMOTE_PART],ANY_HOST, 0);
        }
        else
        {
		key.h = *rh;
  		partPush(p[REMOTE_PART],(char*) &key.h, LOCALHOSTLEN);
        }
  }
  else
  {
	return XK_FAILURE;
  }

  if (mapResolve(((PSTATE *)(s->state))->passivemap, &key, &en)==XK_SUCCESS)
    {
      mapRemoveBinding(((PSTATE *)(s->myprotl->state))->passivemap,en->binding);
      return XK_SUCCESS;
    }
  return XK_FAILURE;
}

static void
ipsecspi_getproc_protl(s)
    XObj s;
{
  xAssert(xIsProtocol(s));
  s->control = ipsecspiControlProtl;
  s->open = ipsecspiOpen;
  s->openenable = ipsecspiOpenEnable;
  s->opendisable = ipsecspiOpenDisable;
  s->opendone = ipsecspiOpenDone;
  s->demux = ipsecspiDemux;
}

static void
ipsecspi_getproc_sessn(s)
    XObj s;
{
  xAssert(xIsSession(s));
  s->push = ipsecspiPush;
  s->pop = ipsecspiPop;
  s->control = ipsecspiControlProtl;
  s->close = ipsecspiCloseSessn;
}
