/*
 * $RCSfile: lcp_config.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: lcp_config.c,v $
 * Revision 1.3  1994/10/28 21:02:35  andrey
 * *** empty log message ***
 *
 * Revision 1.2  1994/10/27  17:27:04  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:46:02  andrey
 * Initial revision
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/10/28 21:02:35 $
 */

/*
 * options processing routines
 */

/*
 * dispense magic numbers
 */
static u_long
lcp_magic()
{
  return 0;
}

/*
 * configuration option processing
 * - still need to handle variably sized AUTH and QUAL properly
 */

/*
 * see if peer's parameters are ok
 * return anything unpalatable
 * need to check if we've rejected any of these previously
 */
static xkern_return_t
options_acceptable(
  XObj sess,
  PPP_options * lo,
  PPP_options ** rej_opts,
  PPP_options ** nak_opts,
  int rej_not_nak)
{
  opt_list_elem * ole = 0;
  *rej_opts = 0;
  *nak_opts = 0;

  ole = first_option(lo);
  while (ole) {
    switch(ole->type) {

      case LCP_MRU:
        {
          MRU_opt * mp = (MRU_opt *)ole->opt_ptr;
          if (mp->mru < LCP_MIN_MRU) {
            if (rej_not_nak) {
              if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
                ppp_add_option(*rej_opts,
                                lcp_duplicate_option(ole->opt_ptr, ole->type),
                                ole->type);
            }
            /* if we haven't rejected anything yet */
            else if (!*rej_opts) {
              mp = X_NEW(MRU_opt);
              mp->mru = LCP_MIN_MRU;
              if (!*nak_opts) *nak_opts = X_NEW(PPP_options);
              ppp_add_option(*nak_opts, mp, ole->type);
            }
          }
        }
      break;

      case LCP_MAGIC:
        {
          MAGIC_opt * mopt = (MAGIC_opt *)ole->opt_ptr;
          /* hmmm.... */
          if (mopt->magic_number == ((SSTATE*)sess->state)->magic_number) {
            if (rej_not_nak) {
              if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
                ppp_add_option(*rej_opts,
                           lcp_duplicate_option(ole->opt_ptr, ole->type),
                                ole->type);
            }
            /* if we haven't rejected anything yet */
            else if (!*rej_opts) {
              mopt = X_NEW(MAGIC_opt);
              mopt->magic_number = lcp_magic();
              if (!*nak_opts) *nak_opts = X_NEW(PPP_options);
              ppp_add_option(*nak_opts, mopt, ole->type);
            }
          }
        }
      break;

      case LCP_ACCM:
        {
	  /*
           * need to save peer's EscCharMap
           */
          if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
          ppp_add_option(*rej_opts,
                         lcp_duplicate_option(ole->opt_ptr, ole->type),
                         ole->type);
        }
      break;

      case LCP_AUTH:
        {
          if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
          ppp_add_option(*rej_opts,
                         lcp_duplicate_option(ole->opt_ptr, ole->type),
                     ole->type);
        }
      break;

      case LCP_QUAL:
          if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
          ppp_add_option(*rej_opts,
                         lcp_duplicate_option(ole->opt_ptr, ole->type),
                         ole->type);
      break;

      /* reject outright */
      case LCP_ACFC:
      case LCP_PFC:
          if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
          ppp_add_option(*rej_opts,
                         lcp_duplicate_option(ole->opt_ptr, ole->type),
                         ole->type);
      break;

      case LCP_UNKNOWN: /* reject it straight away */
        if (!*rej_opts) *rej_opts = X_NEW(PPP_options);
        ppp_add_option(*rej_opts,
                       lcp_duplicate_option(ole->opt_ptr, ole->type),
                       ole->type);
      break;
    }
    ole = next_option(ole);
  }
  if (*rej_opts || *nak_opts) {
    if (*rej_opts && *nak_opts) {
       ppp_destroy_options(*nak_opts, lcp_destroy_opt);
       nak_opts = 0;
    }
    return XK_FAILURE;
  }
  return XK_SUCCESS;
}

/*
 * recalculate offered nak'd options
 */
static PPP_options *
LCP_options_reconsider(XObj sess, PPP_options * lo)
{
  PPP_options * retval = X_NEW(PPP_options);
  opt_list_elem * ole = 0;

  PPP_options_init(retval);

  ole = first_option(lo);
  while (ole) {
    switch(ole->type) {
      case LCP_MRU:
      break;
      case LCP_ACCM:
      break;
      case LCP_AUTH:
      break;
      case LCP_QUAL:
      break;
      case LCP_MAGIC:
      break;
      case LCP_PFC:
      break;
      case LCP_ACFC:
      break;
    }
    ole = next_option(ole);
  }
  return retval;
}

static void
LCP_opt_store(void * src, char * dest, long count, void * type)
{
  int lcp_t = (int) type;
  u_short s;
  u_long l;

  switch(lcp_t) {
    case LCP_MRU:
      s = htons(((MRU_opt*)src)->mru);
      bcopy((char *)&s, (char *)&((MRU_opt*)dest)->mru, sizeof(s));
    break;

    case LCP_ACCM:
      l = htonl(((ACCM_opt*)src)->accm);
      bcopy((char *)&l, (char *)&((ACCM_opt*)dest)->accm, sizeof(l));
    break;

    case LCP_AUTH:
      s = htons(((AUTH_opt*)src)->auth_prot);
      bcopy((char *)&s, (char *)&((AUTH_opt*)dest)->auth_prot, sizeof(s));
    break;

    case LCP_QUAL:
      s = htons(((QUAL_opt*)src)->quality_prot);
      bcopy((char *)&s, (char *)&((QUAL_opt*)dest)->quality_prot, sizeof(s));
    break;

    case LCP_MAGIC:
      l = htonl(((MAGIC_opt*)src)->magic_number);
      bcopy((char *)&l, (char *)&((MAGIC_opt*)dest)->magic_number, sizeof(l));
    break;

    default:
      bcopy(src, (char *)dest, count);
    break;
  }
}

static long
LCP_opt_load(void * dest, char * src, long count, void * type)
{
  int lcp_t = (int) type;
  int copied = -1;
  u_short s;
  u_long l;

  switch(lcp_t) {
    case LCP_MRU:
      bcopy((char *)&((MRU_opt*)src)->mru, (char *)&s, sizeof(s));
      ((MRU_opt*)dest)->mru = ntohs(s);
      copied = sizeof(s);
    break;

    case LCP_ACCM:
      bcopy((char *)&((ACCM_opt*)src)->accm, (char *)&l, sizeof(l));
      ((ACCM_opt*)dest)->accm = ntohl(l);
      copied = sizeof(l);
    break;

    case LCP_AUTH:
      bcopy((char *)&((AUTH_opt*)src)->auth_prot, (char *)&s, sizeof(s));
      ((AUTH_opt*)dest)->auth_prot = ntohs(s);
      copied = sizeof(s);
    break;

    case LCP_QUAL:
      bcopy((char *)&((QUAL_opt*)src)->quality_prot, (char *)&s, sizeof(s));
      ((QUAL_opt*)dest)->quality_prot = ntohs(s);
      copied = sizeof(s);
    break;

    case LCP_MAGIC:
      bcopy((char *)&((MAGIC_opt*)src)->magic_number, (char *)&l, sizeof(l));
      ((MAGIC_opt*)dest)->magic_number = ntohl(l);
      copied = sizeof(l);
    break;

    default:
      bcopy(src, (char *)dest, count);
      copied = count;
  }
  return copied;
}

/*
 * turns an LCP_options list object into a Msg
 */
static Msg *
make_config_msg(PPP_options * this)
{
  Msg * new_msg = X_NEW(Msg);
  PPP_opt lo;
  opt_list_elem * ole;
  int size = 0;

  msgConstructEmpty(new_msg);
  ole = first_option(this);

  while (ole) {
    switch(ole->type) {
      case LCP_MRU:
        size = sizeof(MRU_opt);
        MRU_opt_init(&lo);
      break;
      case LCP_ACCM:
        size = sizeof(ACCM_opt);
        ACCM_opt_init(&lo);
      break;
      case LCP_AUTH:
        size = sizeof(AUTH_opt);
        AUTH_opt_init(&lo);
      break;
      case LCP_QUAL:
        size = sizeof(QUAL_opt);
        QUAL_opt_init(&lo);
      break;
      case LCP_MAGIC:
        size = sizeof(MAGIC_opt);
        MAGIC_opt_init(&lo);
      break;
      case LCP_PFC:
        size = 0;
        PFC_opt_init(&lo);
      break;
      case LCP_ACFC:
        size = 0;
        ACFC_opt_init(&lo);
      break;

      case LCP_UNKNOWN:
        {
          UNKNOWN_opt * uop = (UNKNOWN_opt *)ole->opt_ptr;
          xTrace1(lcpp, TR_SOFT_ERRORS,
                "make_msg: odd configuration option %d", uop->type);
          lo.type = uop->type;
          lo.length = uop->length;
          msgAppend(new_msg, PPP_opt_hdr_store, &lo, sizeof(PPP_opt), 0, 1);
          if (lo.length) {
            msgAppend(new_msg, LCP_opt_store, uop->buf, lo.length,
                    (void*)LCP_UNKNOWN, 1);
          }
          goto skip;
        }
    }
    msgAppend(new_msg, PPP_opt_hdr_store, &lo, sizeof(PPP_opt), 0, 1);
    if (size) {
      msgAppend(new_msg, LCP_opt_store, ole->opt_ptr, size,
                (void *)((int)ole->type), 1);
    }

skip:
    ole = next_option(ole);
  }
  return new_msg;
}

/*
 * turns a Msg object into a options object, preserving order
 */
static PPP_options *
make_config_options(Msg * this, int len)
{
  PPP_options * new_options = X_NEW(PPP_options);
  PPP_options_init(new_options);

  while(len > 0) {
    PPP_opt lo;
    void * o;

    /* return null if we can't tear down packet? */
    if (msgPop(this, PPP_opt_hdr_load, &lo, sizeof(PPP_opt), 0) == FALSE) {
       ppp_destroy_options(new_options, lcp_destroy_opt);
       return 0;
    } 

    len -= sizeof(PPP_opt);

    switch(lo.type) {
      case LCP_MRU:
        o = X_NEW(MRU_opt);
        if (msgPop(this, LCP_opt_load, o, sizeof(MRU_opt), (void *) LCP_MRU)
                                                                   == FALSE) {
          ppp_destroy_options(new_options, lcp_destroy_opt);
          return 0;
        }
        len -= sizeof(MRU_opt);
      break;

      case LCP_ACCM:
        o = X_NEW(ACCM_opt);
        if (msgPop(this, LCP_opt_load, o, sizeof(ACCM_opt), (void *) LCP_ACCM)
                                                                   == FALSE) {
          ppp_destroy_options(new_options, lcp_destroy_opt);
          return 0;
        }
        len -= sizeof(ACCM_opt);
      break;

      case LCP_AUTH:
        o = X_NEW(AUTH_opt);
        if (msgPop(this, LCP_opt_load, o, sizeof(AUTH_opt), (void *) LCP_AUTH)
                                                                   == FALSE) {
          ppp_destroy_options(new_options, lcp_destroy_opt);
          return 0;
        }
        len -= sizeof(AUTH_opt);
      break;

      case LCP_QUAL:
        o = X_NEW(QUAL_opt);
        if (msgPop(this, LCP_opt_load, o, sizeof(QUAL_opt), (void *) LCP_QUAL)
                                                                   == FALSE) {
          ppp_destroy_options(new_options, lcp_destroy_opt);
          return 0;
        }
        len -= sizeof(QUAL_opt);
      break;

      case LCP_MAGIC:
        o = X_NEW(MAGIC_opt);
        if (msgPop(this, LCP_opt_load, o, sizeof(MAGIC_opt), (void *) LCP_MAGIC)
                                                                   == FALSE) {
          ppp_destroy_options(new_options, lcp_destroy_opt);
          return 0;
        }
        len -= sizeof(MAGIC_opt);
      break;

      case LCP_PFC:
      case LCP_ACFC:
        o = 0;
      break;

      default:
        {
          UNKNOWN_opt * uop;
          xTrace1(lcpp, TR_SOFT_ERRORS,
                "make_config_options: odd configuration option %d", lo.type);
          uop = o = X_NEW(UNKNOWN_opt);
          uop->type = lo.type;
          uop->length = lo.length;
          if (lo.length) {
            uop->buf = xMalloc(lo.length);
          }
          else {
            uop->buf = 0;
          }
          msgPop(this, LCP_opt_load, uop->buf, lo.length, (void *)LCP_UNKNOWN);
          ppp_add_option(new_options, o, LCP_UNKNOWN);
          len -= lo.length;
          goto skip;
        }
    }
    ppp_add_option(new_options, o, lo.type);

skip:
    ;
  }
  return new_options;
}

/*
 * initialize magic numbers
 */
static void
lcp_magic_init()
{
}

