/*
 * $RCSfile: sink.c,v $
 *
 * The SINK protocol absorbs all incomming datagrams from both above
 * and below.  In this manner it functions as a data sink.  Control
 * operations are however always passed on to any lower level protocol
 * if one exists, otherwise an error is returned.
 *
 * This is designed to be used during debugging to temporarily disable
 * certain branches of a graph while retaining any critical control
 * operations.  In this manner a branch simply appears inactive rather
 * than broken.
 *
 * Copyright (c) 1996	Arizona Board of Regents
 *
 * $Log: sink.c,v $
 * Revision 1.2  1997/03/10 22:11:28  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/11/04  23:41:30  mjk
 * Initial revision
 *
 *
 */


#include "xkernel.h"


void			sink_init(XObj);
static xmsg_handle_t	sink_push(XObj, Msg *);
static int		sink_control(XObj, int, char *, int);
static xkern_return_t	sink_demux(XObj, XObj, Msg *);
static xkern_return_t	sink_openenable(XObj, XObj, XObj, Part *);

int	tracesinkp;


/* 
 *  Public Functions
 */

void
sink_init(XObj self) 
{
  XObj	llp;
  Part	part;

  xTraceP0(self, TR_GROSS_EVENTS, "init");
  
  if ( xIsProtocol(llp = xGetDown(self, 0)) ) {
    partInit(&part, 1);
    partPush(part, ANY_HOST, 0);
    if ( xOpenEnable(self, self, llp, &part) == XK_FAILURE ) {
      xError("sink cannot openenable lower protocol");
      return;
    }
  }
  
  self->openenable  = sink_openenable;
  self->control	    = sink_control;
  self->push	    = sink_push;
  self->demux       = sink_demux;
  self->up	    = NULL;
} /* sink_init */



/* 
 * Public Protocol Methods 
 */

static xkern_return_t
sink_openenable(XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
				/* No protection against multiple
                                   openenables. */
  xTraceP0(self, TR_MAJOR_EVENTS, "openenable");
  self->up = hlpRcv;
  return XK_SUCCESS;
} /* sink_openenable */


static int
sink_control(XObj self, int opcode, char *buf, int buflen)
{
  XObj	llp = xGetDown(self, 0);

  xTraceP0(self, TR_MAJOR_EVENTS, "control");
  if ( xIsProtocol(llp) )
    return xControl(llp, opcode, buf, buflen);
  return 0;
} /* sink_control */


static xmsg_handle_t
sink_push(XObj self, Msg *packet)
{
  xTraceP0(self, TR_MAJOR_EVENTS, "push");
  return  XMSG_NULL_HANDLE;
} /* sink_push */

static xkern_return_t
sink_demux(XObj self, XObj lls, Msg *packet)
{
  xTraceP0(self, TR_MAJOR_EVENTS, "demux");
  return XK_SUCCESS;
} /* bounce_demux */


/* sink.c ends here */
 


