/*#include <sys/systeminfo.h>*/
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include "xkernel.h"
#include "tcp.h"
#include "ip.h"
#include "udp.h"
#include "x_stdio.h"
#include "tcpf_internal.h"

#define msgsQueued( cs )    ((cs)->head.next != &((cs)->head))

typedef enum { 
  TCPF_INACTIVE=0, TCPF_OPENING, TCPF_FLUSHING, TCPF_OK, TCPF_QUEUED,
  TCPF_NOT_OPENED
} ConnectStatus;

#define HOST_TYPE IPhost
#define BUFFER_SIZE 4096

#define TRACE_VAR tcpf_rp
#define TCP_NAME "tcp"
#define MAP_SIZE 101
#define INIT_SEQ 1
#define PRINT_MSG 0
#define QUEUE_INTERVAL   500   /* delay for queueTimeout, in ms */
#define LOST_MSG_TIMEOUT (20 * 1000000)  /* delay for lost msg, in us */
#define CLOSE_TICKS      240    /* number of QUEUE_INTERVALs before close */
#define MSG_SEQ_CUTOFF   10     /* buffer messages less than this, even if
				 * no open message received yet
				 */
#define WAIT_FOR_OPEN    (5 * 1000000)   /* time to wait for open if data
					  * arrives first, in us
					  */
typedef struct {
  Protl tcp_protl, link_protl;
  Sessn link_sessn;
  Map   bindmap;
  u_int next_CID;
} PState;

typedef struct msgnode *MsgQ;

struct msgnode {
  MsgQ       next, prev;
  Msg        *msg;            /* separate msg from node */
  u_int      seq;
};

#define TCPF_CLOSE_RCVD  (1<<0)

typedef struct {
  ConnectStatus  status;
  u_char         flags;
  PState         *ps;
  IPhost         SourceAddr, DestAddr;
  long           SourcePort, DestPort;
  u_int          CID, seq;
  Sessn          lls;
  struct msgnode head;
  u_int          nextInQ;
  Event          timeoutEvent, queueEvent, noOpenEvent;
  u_int          CloseAtSeq;
  int            CloseCounter;
} CState;

int tracetcpf_rp;

static xkern_return_t   tcpf_rDemux( Protl, Sessn, Msg * );
static xkern_return_t   openDone( Protl, Protl, Sessn, Protl );
static xkern_return_t   closeDone( Sessn );
static xkern_return_t   customDemux( Protl, Sessn, Msg * );
static xkern_return_t   customOpenDone( Protl, Protl, Sessn, Protl );
static void             identifyDownProtocols ( Protl );
static void             handleOpenMsg ( Protl, Msg * );
static void             handleDataMsg ( Protl, Msg * );
static void             handleCloseMsg ( Protl, Msg * );
static void             int2IPhost ( int, IPhost * );
static void             QueueMsg ( CState *, Msg *, u_int );
static Msg              *DequeueMsg ( CState *);
/*void printBuf ( char *ptr, long len );*/
/*static void printMessage ( Msg *dg );*/
void flushMsgQueue (CState *cs);
void DumpQueue (CState *cs);
void QueueTimeout (Event e, void *arg);
void CloseConnection (CState *cs);
void LostMsgTimeout (Event e, void *arg);
void noOpenTimeout (Event e, void *arg);

int tcpf_r_init (Protl);

int tcpf_r_init (self)
     Protl self;
{
  PState	*ps;
  Part p;
  
  printf ("I am the receive-side router\n");
  
  /* Initialize protocol method table */
  self->demux = tcpf_rDemux;
  self->opendone = openDone;
  self->closedone = closeDone;

  ps = X_NEW(PState);
  bzero((char *)ps, sizeof(PState));
  self->state = (VOID *)ps;

  identifyDownProtocols (self);
  /*  tcpfProcessRomFile(self); */
  str2ipHost(&tcpfRcvrAddr, "192.12.71.2");
  tcpfRcvrPort = 999;
  
  ps->next_CID = rand();
  ps->bindmap = mapCreate (MAP_SIZE, sizeof (u_int));
  
  /* Do xOpenEnable on link protocol */
  partInit (&p, 1);
  partPush(p, &(tcpfRcvrAddr), sizeof(HOST_TYPE));
  partPush(p, &(tcpfRcvrPort), sizeof(LinkPort));
  
  if ( xOpenEnable(self, self, ps->link_protl, &p) == XK_FAILURE ) 
    {
      Kabort ("TCPF_R can't openenable link protocol");
    } 

  return 0;
}


static void identifyDownProtocols (self)
     Protl self;
{
  Protl llp1, llp2, tcp;
  PState	*ps = (PState *)self->state;
  
  xTrace1(TRACE_VAR, TR_EVENTS, "Using \"%s\" for TCP name", TCP_NAME);

  tcp = xGetProtlByName (TCP_NAME);

  if (!xIsProtocol(tcp)) {
      Kabort ("Could not find a protocol object for TCP");
    }
  
  llp1 = xGetDown(self, 0);
  llp2 = xGetDown(self, 1);
  if (!xIsProtocol(llp1) || !xIsProtocol(llp2)) {
      Kabort("TCPF_R does not have two lower protocols");
    }

  if (llp1 == tcp)
    {
      ps->tcp_protl = llp1;
      ps->link_protl = llp2;
    }
  else if (llp2 == tcp)
    {
      ps->link_protl = llp1;
      ps->tcp_protl = llp2;
    }
  else
    {
      Kabort("TCPF_R is not configured over TCP");
    }
}


static
#if defined(__GNUC__)
inline
#endif
void checkHandle( h, str )
    xmsg_handle_t	h;
    char		*str;
{
    switch ( h ) {
      case XMSG_ERR_HANDLE:
      case XMSG_ERR_WOULDBLOCK:
	sprintf(errBuf, "%s returns error handle %d", str, h);
	Kabort(errBuf);
      default:
	;
    }
}

/*
void printBuf ( char *ptr, long len )
{
  char buf[len + 1];

  memcpy (buf, ptr, len);

  buf[len] = '\0';
  printf("%s", buf);
}


static void printMessage (dg)
     Msg *dg;
{
  MsgWalk ctx;
  int  len;
  char *src;

  if (msgLen(dg) == 0)
    return;

  msgWalkInit(&ctx, dg);
  while ((src = msgWalkNext(&ctx, &len)) != 0) 
    {
      printBuf (src, len);
    }

  msgWalkDone(&ctx);
}
*/

/*
 * Receive a message from one-way link and forward it out on the
 * appropriate TCP connection.
 */
xkern_return_t
tcpf_rDemux( self, lls, dg )
    Protl self;
    Sessn lls;
    Msg *dg;
{
  PState	*ps = (PState *)self->state;
  static int c = 1;
  char *buf;
  int  type;

  if (lls != ps->link_sessn)
    {
      return XK_FAILURE;
    }
  
  /* Demux on message type.
   * Unfortunately, "buf" does not always seem to be 4-byte aligned,
   * necessitating a copy.
   */
  buf = msgPeek(dg, 4, FALSE);
  memcpy(&type, buf, 4);
  
  switch (ntohl(type))
    {
    case OPEN_MSG:
      handleOpenMsg (self, dg);
      break;
      
    case DATA_MSG:
      /* printf ("msgLen = %d\n", msgLen(dg)); */
      handleDataMsg (self, dg);
      break;
      
    case CLOSE_MSG:
      handleCloseMsg (self, dg);
      break;
      
    default:
      fprintf (stderr, "ERROR: Unknown message type in Demux\n");
      break;
    }

  if (PRINT_MSG)
    {
      /*      printMessage (dg); */
      return XK_SUCCESS;
    }
  else
    xIfTrace(TRACE_VAR, TR_MAJOR_EVENTS) 
      {
	putchar('.');
	if (! (c++ % 50)) 
	  {
	    putchar('\n');
	  }
	return XK_SUCCESS;
      } 

  return XK_SUCCESS;
  
}


static void 
handleOpenMsg ( self, dg )
     Protl self;
     Msg *dg;
{
  PState	*ps = (PState *)self->state;
  CState *cs;
  struct OpenHdr hdr;
  char  *buf;
  Part  p[2];
  int   noblock = 1;
  int   CID;
  
  memcpy(buf + sizeof(struct OpenHdr), dg, sizeof(Msg));
  memcpy(&hdr, buf, sizeof(struct OpenHdr));

  CID = ntohl(hdr.CID);
  if (mapResolve(ps->bindmap, &CID, (void **)&cs) == XK_FAILURE) {
      cs = X_NEW(CState);
      bzero ((char *) cs, sizeof(CState));
      
      cs->CID = CID;
      cs->head.next = cs->head.prev = &(cs->head);
      cs->seq = ntohl(hdr.seq) + 1;
      if ( mapBind (ps->bindmap, (void *) &(cs->CID), (void *) cs) == ERR_BIND)
	{
	  xTrace0 (TRACE_VAR, TR_ERRORS, 
		   "handleOpenMsg: binding failed");
	  return ;
	}
    }
  else
    {
      /* Connection state already exists */
      if (cs->status != TCPF_NOT_OPENED)
	{
	  xTrace0 (TRACE_VAR, TR_ERRORS, 
		   "handleOpenMsg: received open for existing connection");
	  return;
	}
      cs->seq++;
      evCancel (cs->noOpenEvent);
      cs->noOpenEvent = NULL;
    }

  /*
   * Initialize the connection state 
   */
  cs->status = TCPF_OPENING;
  int2IPhost (hdr.SourceAddr, &cs->SourceAddr);
  cs->SourcePort = ntohl (hdr.SourcePort);
  int2IPhost (hdr.DestAddr, &cs->DestAddr);
  cs->DestPort = ntohl (hdr.DestPort);
  cs->ps = ps;
  cs->timeoutEvent = NULL;
  cs->CloseCounter = 0;
  
  xTrace1 (TRACE_VAR, TR_EVENTS, "\nOpening connection, CID %d", cs->CID);
  xTrace5 (TRACE_VAR, TR_EVENTS, 
	   "Source IP addr: %d.%d.%d.%d, TCP port: %d",
	   cs->SourceAddr.a, cs->SourceAddr.b, cs->SourceAddr.c, 
	   cs->SourceAddr.d, cs->SourcePort);
  xTrace5 (TRACE_VAR, TR_EVENTS, "Dest IP addr: %d.%d.%d.%d, TCP port: %d",
	   cs->DestAddr.a, cs->DestAddr.b, cs->DestAddr.c, 
	   cs->DestAddr.d, cs->DestPort);
  
  /* Open TCP connection with destination, save pointer to lls */
  partInit (p, 2);
  partPush (p[0], &(cs->DestAddr), sizeof(IPhost));
  partPush (p[0], &(cs->DestPort), sizeof(long));
  partPush (p[1], &(cs->SourceAddr), sizeof(IPhost));
  partPush (p[1], &(cs->SourcePort), sizeof(long));

  cs->lls = xOpen (self, self, ps->tcp_protl, p);
  if ( cs->lls == ERR_SESSN ) 
    {
      xTrace0 (TRACE_VAR, TR_ERRORS, "Could not open TCP connection");
      /* Take appropriate action here */
      cs->status = TCPF_INACTIVE;
      printf ("ERROR: Unable to open TCP connection!!!\n");
      CloseConnection (cs);
      return;
    }

  xControl(cs->lls, SETNONBLOCKINGIO, (char *) &noblock, sizeof(int));
  cs->queueEvent = evSchedule (QueueTimeout, cs, 0);
  
  /* 
   * flushMsgQueue() takes care of state changes, so call it even if
   * no messages are queued.
   */
  flushMsgQueue(cs);
}


void flushMsgQueue (CState *cs)
{
  Msg *m;
  xkern_return_t status;

  if (cs->status == TCPF_NOT_OPENED)
    return;
  
  cs->status = TCPF_FLUSHING;
  
  while (msgsQueued(cs) && cs->head.next->seq == cs->seq)
    {
      m = DequeueMsg (cs);
      status = xPush (cs->lls, m);
      if (status == XMSG_ERR_WOULDBLOCK)
	{
	  QueueMsg (cs, m, cs->seq);
	  xTrace1 (TRACE_VAR, TR_EVENTS, "Message %d would block -- requeued",
		   cs->seq);
	  break;
	}
      msgDestroy (m);
      xTrace1 (TRACE_VAR, TR_EVENTS, "Sent %d from queue", cs->seq);
      cs->seq++;
    }

  if (msgsQueued(cs) && cs->head.next->seq == cs->seq)
    cs->status = TCPF_QUEUED;
  else
    {
      cs->status = TCPF_OK;

      /* If we've found a hole in the queue sequence, schedule timeout */
      if (msgsQueued(cs) && cs->timeoutEvent == NULL)
	cs->timeoutEvent = evSchedule (LostMsgTimeout, cs, LOST_MSG_TIMEOUT);
      
      if ((cs->flags & TCPF_CLOSE_RCVD) && (cs->CloseAtSeq == cs->seq))
	CloseConnection (cs);
    }
}


static void 
handleDataMsg ( self, dg )
     Protl self;
     Msg *dg;
{
  PState	*ps = (PState *)self->state;
  CState *cs;
  struct DataHdr hdr;
  char  *buf;
  xkern_return_t status;
  int CID;
  
  memcpy(buf + sizeof(struct DataHdr), dg, sizeof(Msg));
  memcpy(&hdr, buf, sizeof(struct DataHdr));

  hdr.CID = ntohl(hdr.CID);
  hdr.seq = ntohl(hdr.seq);
  
  /* Retrieve state from map based on CID */
  if (mapResolve(ps->bindmap, &(hdr.CID), (void **)&cs) == XK_FAILURE) {
      if (hdr.seq > MSG_SEQ_CUTOFF)
	{
	  xTrace0 (TRACE_VAR, TR_ERRORS, 
		   "received Data message belongs to no known connection");
	  return;
	}
      else
	{
	  /*
	   * Create state and queue message
	   */
	  printf ("Got some data before Open message!\n");
	  
	  cs = X_NEW(CState);
	  bzero ((char *) cs, sizeof(CState));
      
	  cs->CID = CID;
	  cs->head.next = cs->head.prev = &(cs->head);
	  cs->status = TCPF_NOT_OPENED;
	  /*
	   * Currently, the OPEN message has sequence number 1.
	   */
	  cs->seq = 1;   
	  
	  if ( mapBind (ps->bindmap, (void *) &(cs->CID), (void *) cs) 
	      == ERR_BIND)
	    {
	      xTrace0 (TRACE_VAR, TR_ERRORS, 
		       "handleDataMsg: binding failed");
	      return ;
	    }
	  cs->noOpenEvent = evSchedule (noOpenTimeout, cs, WAIT_FOR_OPEN);
	}
    }
  
  /*
   * Drop this message if the state is found but is inactive
   */
  if (cs->status == TCPF_INACTIVE)
    return;
  
  if (hdr.seq == 0)
    {
      xTrace0 (TRACE_VAR, TR_MAJOR_EVENTS, "Sequence space wrapped!");
    }

  /* reset CloseCounter (detects an idle connection) */
  cs->CloseCounter = 0;
  
  /* Check sequence number */
  if (cs->seq != hdr.seq)
    {
      if (LT(cs->seq, hdr.seq))
	{
	  xTrace1 (TRACE_VAR, TR_EVENTS,  "Got %d -> queued.",
	       hdr.seq);
	  QueueMsg (cs, dg, hdr.seq);
	  xIfTrace(TRACE_VAR, TR_EVENTS)
	    {
	      DumpQueue (cs);
	    }
	  return;
	}
      else
	{
	  xTrace1 (TRACE_VAR, TR_EVENTS, 
	       "Got %d -> dropped", hdr.seq);
	  /* it's a duplicate message */
	  return;
	}
    }

  /* If there is a timeout scheduled, it must be for this message */
  if (cs->timeoutEvent != NULL)
    {
      evCancel (cs->timeoutEvent);
      cs->timeoutEvent = NULL;
    }
  
  switch (cs->status)
    {
    case TCPF_OPENING:
    case TCPF_FLUSHING:
      /* queue message for later transmission */
      xTrace1 (TRACE_VAR, TR_EVENTS,  "Got %d -> queued.",
	       hdr.seq);
      QueueMsg (cs, dg, hdr.seq);
      xIfTrace(TRACE_VAR, TR_EVENTS)
	{
	  DumpQueue (cs);
	}
      
      break;

    case TCPF_QUEUED:
      /* 
       * Message to go out is at the front of the queue.
       * Queue this message, and then try to flush the queue.
       */
      xTrace1 (TRACE_VAR, TR_EVENTS,  "Got %d -> queued.",
	       hdr.seq);
      QueueMsg (cs, dg, hdr.seq);
      flushMsgQueue (cs);
      break;
      
    case TCPF_OK:
      xTrace1 (TRACE_VAR, TR_EVENTS, "Got %d -> sending", hdr.seq);
      status = xPush (cs->lls, dg);
      if (status == XMSG_ERR_WOULDBLOCK)
	{
	  QueueMsg (cs, dg, hdr.seq);
	  xTrace1 (TRACE_VAR, TR_EVENTS, "Message %d would block -- queued",
		   hdr.seq);
	  cs->status = TCPF_QUEUED;
	  break;
	}
      cs->seq++;
      if (cs->status == TCPF_OK)
	flushMsgQueue (cs);
      break;

    default:
      ;
    }
}


static void 
handleCloseMsg ( self, dg )
     Protl self;
     Msg *dg;
{
  PState	*ps = (PState *)self->state;
  CState *cs;
  struct CloseHdr hdr;
  char   *buf;
  
  memcpy(buf + sizeof(struct CloseHdr), dg, sizeof(Msg));
  memcpy(&hdr, buf, sizeof(struct CloseHdr));
  hdr.CID = ntohl(hdr.CID);
  hdr.seq = ntohl(hdr.seq);
  
  printf ("\nGot a close message for CID %d\n", hdr.CID);

  /* Retrieve state from map based on CID */
  if (mapResolve(ps->bindmap, &(hdr.CID), (void **)&cs) == XK_FAILURE) {
      xTrace0 (TRACE_VAR, TR_ERRORS, 
	       "received Close message belongs to no known connection");
      return;
    }
  
  /* 
   * Remember that we've seen a close message for this connection, but 
   * don't close until all sequence numbers before this msg have been
   * sent.
   */
  if (cs->status != TCPF_INACTIVE)
    {
      cs->flags |= TCPF_CLOSE_RCVD;
      cs->CloseAtSeq = hdr.seq;
    }

  /* Try to flush the queue and close now */
  if (cs->status == TCPF_OK || cs->status == TCPF_QUEUED)
    flushMsgQueue(cs);
}


void CloseConnection (CState *cs)
{
  Msg *m;
  
  if (cs->status != TCPF_INACTIVE)
    xClose (cs->lls);

  if (cs->queueEvent != NULL)
    {
      evCancel (cs->queueEvent);
      cs->queueEvent = NULL;
    }
  
  m = DequeueMsg (cs);
  while (m != NULL)
    {
      msgDestroy (m);
      m = DequeueMsg (cs);
    }

  /* Remove state from map */
  if ( mapUnbind (cs->ps->bindmap, &(cs->CID)) == XK_FAILURE )
    {
      xTrace0 (TRACE_VAR, TR_ERRORS, "mapRemoveKey failed");
    }

}


static xkern_return_t
openDone( self, llp, s, hlpType )
    Protl self, llp, hlpType;
    Sessn s;
{
  PState	*ps = (PState *)self->state;

  xTrace0(TRACE_VAR, TR_MAJOR_EVENTS, "TCPF_R openDone");

  xDuplicate(s);
  ps->link_sessn = s;
  
  return XK_SUCCESS;
}


static xkern_return_t
closeDone( lls )
    Sessn	lls;
{
  printf ("tcpf_r:  In closeDone\n");
  
  return XK_SUCCESS;
}


static xkern_return_t
customDemux( self, lls, dg )
    Protl	self;
    Sessn	lls;
    Msg		*dg;
{
  u_short space = BUFFER_SIZE;

  xTrace1(TRACE_VAR, TR_DETAILED, "TCP custom demux resets buffer to size %d",
	  BUFFER_SIZE);
  if (xControl(lls, TCP_SETRCVBUFSPACE, (char*)&space, sizeof(space)) < 0) 
    xError("customDemux: TCP_SETRCVBUFSPACE failed");
  
  return XK_SUCCESS;
}


static void int2IPhost (int x, IPhost *y)
{
  y->d = x & 0xff;
  y->c = (x >> 8) & 0xff;
  y->b = (x >> 16) & 0xff;
  y->a = (x >> 24) & 0xff;
}


static void QueueMsg ( cs, m, seq )
     CState *cs;
     Msg    *m;
     u_int  seq;
{
  MsgQ   n = X_NEW(struct msgnode), tmp;
  Msg    *new = X_NEW(Msg);
  
  msgConstructCopy (new, m);

  n->msg = new;
  n->seq = seq;
  
  if ( cs->head.next == &(cs->head) || LT(seq, cs->head.next->seq) )
    tmp = &(cs->head);
  else
    for (tmp = cs->head.prev; tmp != &(cs->head) && GT(tmp->seq, seq); 
	 tmp = tmp->prev);
  
  if (tmp != &(cs->head) && tmp->seq == seq)
    {
      /* duplicate -- drop the message */
      msgDestroy (new);
      free (new);
      free (n);
    }
  else
    {
      n->prev = tmp;
      n->next = tmp->next;
      tmp->next->prev = n;
      tmp->next = n;
    }
}


static Msg *DequeueMsg ( cs )
     CState *cs;
{
  MsgQ   n;
  Msg    *m;
  
  if ((n = cs->head.next) != &(cs->head))
    {
      n->next->prev = &(cs->head);
      cs->head.next = n->next;
      m = n->msg;
      free (n);
      return m;
    }
  
  return NULL;
}


void DumpQueue (CState *cs)
{
  MsgQ   n;
  int    c = 7;

  return;
  
  printf ("Queue: ");
  for (n = cs->head.next; n != &(cs->head); n = n->next)
    {
      c += printf ("%d, ", n->seq);
      if (c > 60)
	{
	  c = 0;
	  putc ('\n', stdout);
	}
    }
  putc ('\n', stdout);
}


void QueueTimeout (Event e, void *arg)
{
  CState *cs = (CState *) arg;
  int noblock;
  
  if (evIsCancelled (e))
    return;

  Delay (QUEUE_INTERVAL);

  while (cs->queueEvent != NULL)
    {

      xTrace0 (TRACE_VAR, TR_EVENTS, "Checking message queue...");
/*      
      printf ("Checking message queue...\n");
*/

      /*
       * Closing an idle connection is separate from closing due to a lost
       * message.  Here, we assume all messages are received.  If this is 
       * not the case, the lost message timeout will catch it, not this.
       */
      if (!(cs->flags & TCPF_CLOSE_RCVD))
	{
	  if (++cs->CloseCounter >= CLOSE_TICKS)
	    {
	      xTrace1 (TRACE_VAR, TR_MAJOR_EVENTS, 
		       "Connection %d is idle -- closing", cs->CID);
	      cs->flags |= TCPF_CLOSE_RCVD;
	      if (msgsQueued(cs))
		cs->CloseAtSeq = cs->head.prev->seq + 1;
	      else
		CloseConnection (cs);
	    }
	}
  
      if (cs->status == TCPF_QUEUED)
	{
	  /* 
	   * Temporarily turn off non-blocking I/O -- it doesn't matter if 
	   * this thread blocks.
	   */
	  noblock = 0;
	  xControl(cs->lls, SETNONBLOCKINGIO, (char *) &noblock, sizeof(int));
	  flushMsgQueue (cs);
	  noblock = 1;
	  xControl(cs->lls, SETNONBLOCKINGIO, (char *) &noblock, sizeof(int));
	}
      else
	xTrace1 (TRACE_VAR, TR_EVENTS, "State is %d", cs->status);
      
      Delay (QUEUE_INTERVAL);
    }
  
  evDetach (e);
  xFree ((char *) cs);
}


void LostMsgTimeout (Event e, void *arg)
{
  CState *cs = (CState *) arg;
  
  if (evIsCancelled (e))
    return;

  evDetach(e);
/*  xTrace0 (TRACE_VAR, TR_EVENTS, "Timeout waiting for message!"); */
  printf ("Timed out connection %d waiting for message %d\n", 
	  cs->CID, cs->seq);
  printf ("Closing connection and cleaning up...\n");
  
  CloseConnection (cs);
  
  cs->status = TCPF_INACTIVE;
}

void noOpenTimeout (Event e, void *arg)
{
  CState *cs = (CState *) arg;
  
  if (evIsCancelled (e))
    return;

  evDetach(e);
  xTrace0 (TRACE_VAR, TR_EVENTS, "Timeout waiting for Open message!");
  printf ("Timed out connection %d waiting for Open message\n",
	  cs->CID);
  printf ("Closing connection and cleaning up...\n");
  
  CloseConnection (cs);
  
  cs->status = TCPF_INACTIVE;
}



