/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: xtime.c,v $
 *
 * HISTORY
 * $Log: xtime.c,v $
 * Revision 1.2  1994/06/11 00:02:26  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/04/19  22:51:25  davidm
 * Initial revision
 *
 */
/*
 * Maintains time.
 */
#include "xk_assert.h"
#include "xtime.h"
#include "scout_time.h"

#define PICO_SECS_PER_USEC	(1000L*1000L)
#define PICO_SECS_PER_SEC	(PICO_SECS_PER_USEC*PICO_SECS_PER_USEC)

static u_long	pico_secs_per_tick;
static int	shift = -1;


void
xTimeInit(void)
{
    pico_secs_per_tick = PICO_SECS_PER_SEC / scout_ticks_per_second;
    if ((scout_ticks_per_second & (scout_ticks_per_second - 1)) == 0) {
	/* it's a power of two: */
	shift = ffs(scout_ticks_per_second) - 1;
    } /* if */
} /* xTimeInit */


void
xGetTime(XTime *t)
{
    register u_long now, q, r;

    xAssert(pico_secs_per_tick > 0);

    now = scout_age;
    if (shift >= 0) {
	q = now >> shift;
	r = now - (q << shift);
    } else {
	q = now / scout_ticks_per_second;
	r = now - q * scout_ticks_per_second;
    } /* if */
    t->sec  = q;
    t->usec = (r * pico_secs_per_tick) / PICO_SECS_PER_USEC;
} /* xGetTime */

			/*** end of xtime.c ***/
