/*
 * $RCSfile: usit_main.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.2 $
 * $Date: 1996/02/15 16:29:56 $
 */

#include "usc.h"
#include "usit.h"
#include "symbol.h"
#include "buffer.h"
#include "gen_utils.h"
#include "usit_version.h"

#ifndef OPTIMIZE
extern void print_type_table (int level);
extern void print_sym_table (int level);

extern void print_struct_table (int level);
extern void print_union_table (int level);

#endif

extern void check_config (void);
extern void print_config (int level);
extern int read_config (char *config_file);
extern void init_symbols (void);
extern void init_codegen (void);
extern void finish_all_syms (void);
extern int yydebug;
extern FILE *yyin;

UsitMode usit_output, usit_input;
char *infer_file_name;
char *native_compiler;
char *UsitDataFile;
FILE *UsitData;

char *UsitValidateFile;
FILE *UsitValidate;

Buffer *FunctionsBuffer;

static char *HelpMsg[] = {
  "Command line options:\n",
  "-C native_compiler  -- select native compiler",
  "-D debug_level      -- select amount of debug output",
  "-V                  -- print version",
  "-h                  -- print this message",
  "-i                  -- just output the inference code",
  "-u                  -- output the final usc code",
  "-v                  -- generate code to run validation",
#ifdef YYDEBUG
  "-t                  -- print out yacc trace output",
#endif
  "-o filename         -- select output file",
  "-d filename         -- select a inference data file (can't use with -i)",
  "",
  "Default options are: -C\"gcc -O\" -u -o out.usc",
  NULL
};

/***************************************************************************/
/*                                                                         */
/*  Here's some stubs for functions called in symbol.c and grammar.y       */
/*  but not used by USIT.                                                  */
/*                                                                         */
/***************************************************************************/


void gen_sym_def (Symbol *sym)
{
}

Expr *make_stub (Symbol *function)
{
  return NULL;
}

/***************************************************************************/

void usit_add_function (Buffer *source)
{
  catbuf (FunctionsBuffer, source);
}

void usit_output_functions (FILE *out)
{
  if (BUFFER_LEN (FunctionsBuffer) > 0)
    fprintf (out, "%s\n\n", BUFFER_STRING (FunctionsBuffer));
}

char *allocate_new_string (char *string)
{
  char *new;
  int temp;
  if (string != NULL) {
    temp = strmem (string);
    new = (char *)Malloc (temp * sizeof (char));
    dprintf (1000, "Entering allocate_new_string: %s", string);
    strcpy (new, string);
    dprintf (1000, "Exiting allocate_new_string: %s %s", new, string);
  }
  else {
    new = NULL;
  }
  return new;
}

char *allocate_temp_string (char *string)
{
  char *new;
  if (string != NULL) {
    new = (char *)TempMalloc (strmem (string) * sizeof (char));
    dprintf (1000, "Entering allocate_temp_string: %s", string);
    strcpy (new, string);
    dprintf (1000, "Exiting allocate_temp_string: %s %s", new, string);
  }
  else {
    new = NULL;
  }
  return new;
}

int strmem (char *string)
{
  return strlen (string) + 1;
}

void init (void)
{
  init_mem();
  inityy();
  init_symbols();
  init_gen_utils (OutFile);
  InFileName = NULL;
  OutFile = NULL;
  DefaultCodeType = GEN_PLAIN;
  code_type = DefaultCodeType;
  print_comments = 0;
  make_validate = 0;
  speed_test = 0;
  usit_input = USIT_IN_C_CODE;
  usit_output = USIT_OUT_USC_CODE;
  preserve_type_names = 1;            /* talk to dorgival about this */

  FunctionsBuffer = makebuf (128);

#ifdef USE_DMALLOC
  dmalloc_opt (DMALLOC_HEAPSTAT | DMALLOC_HEAPCHECK); 
#endif
}

void finishup(void)
{
  static int done = 0;

  if ( done )
    return;
  else
    done = 1;

  code_type = GEN_PLAIN;
  run_preprocess();

  if (usit_output == USIT_OUT_USC_CODE)
    usit_output_functions (OutFile);
}

void closeup (void)
{
  finishup();

  close_mem();
#ifdef USE_DMALLOC
  print_heap_stats();
#endif
  if (OutFile != stdout)
    fclose (OutFile);
  if (make_validate)
    fclose (UsitValidate);
}

int main(int argc, char **argv)
{
  char *file = NULL;
  int  option;
  extern char* optarg;
  extern int   optind;
  char *options;
  char *progname = argv[0];

#ifdef YYDEBUG
  options = "C:D:Vhid:uo:tv:";
#else
  options = "C:D:Vhid:uo:v:";
#endif

  init();
  do_preprocess = 1;
  make_validate = 0;
  speed_test = 0;
  print_comments = 0;
  native_compiler = NULL;

  InFileName = NULL;
				/* The value of optind is undefined
                                   before getopt is called.  Here we
                                   force it to one so we can skip
                                   argv[0].  This was needed for
                                   Linux. */
  optind = 1;
  
  while (optind < argc) {
    if (argv[optind][0] == '-') {
      option = getopt (argc, argv, options);
      if (option == EOF) {
	break;
      }
      
      switch (option) {
      case 'C':
	native_compiler = allocate_new_string (optarg);
	break;
	
      case 'D':
	Debug_Level = atoi (optarg);
	break;
	
      case 'V':
      {
	char** vers;
	for ( vers=Version; *vers; vers++ ) {
	  fprintf(stderr,"%s\n",*vers);
	}
	exit(0);
      }
	break;
	
      case 'h':
      {
	char **help;
	for (help = HelpMsg; *help; help++) {
	  fprintf (stderr, "%s\n", *help);
	}
	exit (0);
      }
	
      case 'i':
	usit_output = USIT_OUT_INFER_CODE;
	break;

      case 'v':
	if (make_validate) {
	  fprintf (stderr, "%s: Can't use -v twice!\n", progname);
	  exit (1);
	}
	make_validate = 1;
	UsitValidateFile = allocate_new_string (optarg);
	if ((UsitValidate = fopen (UsitValidateFile, "w")) == NULL) {
	  fprintf (stderr, "%s: Can't open: %s\n", UsitValidateFile, progname);
	  exit (1);
	}
	break;
	
      case 'u':
	usit_output = USIT_OUT_USC_CODE;
	break;
	
      case 'd':
	UsitDataFile = allocate_new_string (optarg);
	usit_input = USIT_IN_INFER_OUTPUT;
	if ((UsitData = fopen (UsitDataFile, "r")) == NULL) {
	  fprintf (stderr, "%s: Can't open: %s\n", UsitDataFile, progname);
	  exit (1);
	}
	break;
	
#ifdef YYDEBUG
      case 't':
	yydebug = 1;
	break;
#endif
      
      case 'o':
	if ( OutFile ) {
	  fprintf(stderr, "%s: Two definitions for the output file\n", 
		  progname);
	  exit(1);
	}
	if (!strcmp (optarg, "-"))
	  OutFile = stdout;
	else
	  OutFile = fopen (optarg, "w");
	if (!OutFile) {
	  fprintf (stderr, "%s: Can't open output file: %s\n", 
		   progname, optarg);
	  exit (1);
	}
	break;
      }
    }
    else {
      if (InFileName) {
	fprintf (stderr, "%s: Can't have two input files\n", progname);
	exit (1);
      }
      else {
	InFileName = allocate_new_string (argv[optind]);
	if (strcmp (InFileName, "-")) {
	  yyin = fopen (InFileName, "r");
	  if (!yyin) {
	    fprintf (stderr, "%s: Can't open input file: %s\n", 
		     progname, argv[optind]);
	    exit (1);
	  }
	}
      }
      optind++;
    }
  }
  argv += optind;
  argc -= optind;

  if (argc > 1) {
    fprintf (stderr, "%s: Unknown option\n", progname);
    exit (1);
  }

  if (usit_input == USIT_IN_INFER_OUTPUT &&
      usit_output == USIT_OUT_INFER_CODE) {
    fprintf (stderr, "%s: Can't use -d with -i\n", progname);
    exit (1);
  }

  if (make_validate && usit_output != USIT_OUT_USC_CODE) {
    fprintf (stderr, "%s: -v must be used with -u", progname);
    exit (1);
  }

  if (!native_compiler) {
    native_compiler = DEFAULT_NATIVE_COMPILER;
    if (usit_input == USIT_IN_C_CODE && usit_output == USIT_OUT_USC_CODE)
      fprintf(stderr,"Using default compiler option: %s\n", native_compiler);
  }
    

  if ( !InFileName ) {
    InFileName = "-";
  }


  if ( !OutFile ) {
    if (usit_output == USIT_OUT_INFER_CODE)
      file = "out.c";
    else if (usit_output == USIT_OUT_USC_CODE)
      file = "out.usc";

    OutFile = fopen (file, "w");
    if (!OutFile) {
      fprintf (stderr, "%s: Can't open output file: %s\n", progname, file);
      exit (1);
    }
  }

  init_outfile(USIT_OUTPUT);

  yyparse();

  print_sym_table (200);
  print_type_table (200);
  
  print_struct_table (250);
  print_union_table (250);

  /*
  code_type = GEN_PLAIN;
  run_preprocess();

  if (usit_output == USIT_OUT_USC_CODE)
    usit_output_functions (OutFile);
  */

  closeup();
  return 0;
}
